""" Command Line Script for running gfdlvitals """

import argparse
import glob
import os
import shutil
import subprocess
import tempfile
import gfdlvitals

__all__ = ["arguments", "process_year", "run"]


def arguments(args=None):
    """
    Function to capture the user-specified command line options
    """
    description = """
    Program for generating global mean statistics directly 
    from history tarfile.

    For help, contact John.Krasting@noaa.gov

    """

    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.RawTextHelpFormatter
    )

    parser.add_argument(
        "historydir",
        metavar="HISTORY DIR",
        type=str,
        default=os.getcwd(),
        help="Path to /history directory",
    )

    parser.add_argument(
        "-o",
        "--outdir",
        type=str,
        default="./",
        help="Output directory. Default is current directory",
    )

    parser.add_argument(
        "-m",
        "--modelclass",
        type=str,
        default="CM4",
        help="Model class. Options include CM4 and  ESM2. Default is CM4",
    )

    parser.add_argument(
        "-c",
        "--component",
        type=str,
        default="all",
        help="Comma-separated list of components to process. "
        + "See documentation for full list. Default is all.",
    )

    parser.add_argument(
        "-s",
        "--startyear",
        type=int,
        default=None,
        help="Starting year to process. Default is all years.",
    )

    parser.add_argument(
        "-e",
        "--endyear",
        type=int,
        default=None,
        help="Ending year to process. Default is all years.",
    )

    parser.add_argument(
        "-g",
        "--gridspec",
        type=str,
        default=None,
        help="Path to gridspec tarfile. Used in AMOC calculation. " + "Default is None",
    )

    args = parser.parse_args(args)
    args.historydir = os.path.abspath(args.historydir)
    if args.gridspec is not None:
        args.gridspec = os.path.abspath(args.gridspec)

    return args


def process_year(args, infile):
    """Function to process a single year

    Parameters
    ----------
    args : argparse.parser
        Parsed commmand line arguments
    infile : str, pathlike
        History tar file path
    """

    # -- Set the model year string
    fyear = str(infile.split("/")[-1].split(".")[0])

    # -- Run the main code
    if args.modelclass == "ESM2":
        gfdlvitals.models.ESM2.routines(args, infile)
    elif args.modelclass == "CM4":
        gfdlvitals.models.CM4.routines(args, infile)

    # -- Move results to their final location
    if not os.path.exists(args.outdir):
        os.makedirs(args.outdir)
    for reg in ["global", "nh", "sh", "tropics"]:
        for component in [
            "Land",
            "Atmos",
            "AtmosAer",
            "AeroCMIP",
            "Ocean",
            "Ice",
            "TOPAZ",
            "COBALT",
            "BLING",
            "OBGC",
            "Timing",
        ]:
            if os.path.exists(fyear + "." + reg + "Ave" + component + ".db"):
                if not os.path.exists(
                    args.outdir + "/" + reg + "Ave" + component + ".db"
                ):
                    shutil.copyfile(
                        fyear + "." + reg + "Ave" + component + ".db",
                        args.outdir + "/" + reg + "Ave" + component + ".db",
                    )
                else:
                    gfdlvitals.util.merge.merge(
                        fyear + "." + reg + "Ave" + component + ".db",
                        args.outdir + "/" + reg + "Ave" + component + ".db",
                    )


def run(args):
    """Function to run the command line tool

    Parameters
    ----------
    args : list
        List of arguments to parse
    """
    cliargs = arguments(args)

    cliargs.outdir = os.path.abspath(cliargs.outdir)

    # -- Get a list of history files
    dirlist = sorted(glob.glob(cliargs.historydir + "/*.tar"))

    # -- Apply start and end year limits, if applicable
    infiles = []
    if (cliargs.startyear is not None) or (cliargs.endyear is not None):
        if cliargs.startyear is None:
            cliargs.startyear = -1
        if cliargs.endyear is None:
            cliargs.endyear = 99999
        for f in dirlist:
            yr = int(os.path.basename(f)[0:4])
            if cliargs.startyear <= yr <= cliargs.endyear:
                infiles.append(f)
    else:
        infiles = dirlist

    # -- DMGET the history files
    if shutil.which("dmget") is not None:
        print("Dmgetting files ...")
        subprocess.call(["dmget"] + infiles)
        print("Complete!")

    # -- Make temporary directory to work in
    cwd = os.getcwd()
    tempdir = tempfile.mkdtemp()
    os.chdir(tempdir)

    # -- Split list of components to process
    cliargs.component = cliargs.component.split(",")

    # -- Loop over history files
    for _infile in infiles:
        process_year(cliargs, _infile)

    # -- Clean up
    os.chdir(cwd)
    shutil.rmtree(tempdir)



""" Driver for ESM2 class models """

import tarfile
import warnings

from gfdlvitals import averagers
from gfdlvitals.util.average import generic_driver


__all__ = ["routines"]


def routines(args, infile):
    """Driver routine for ESM2-class models

    Parameters
    ----------
    args : argparse.parser
        Parsed commmand line arguments
    infile : str, pathlike
        History tar file path
    """

    # -- Open the tarfile
    tar = tarfile.open(infile)

    # -- Set the model year string
    fyear = str(infile.split("/")[-1].split(".")[0])
    print("Processing " + fyear)

    # -- Get list of components to process
    comps = args.component

    # -- Atmos
    modules = {
        "atmos_month": "Atmos",
        "atmos_level": "Atmos",
    }
    if any(comp in comps for comp in ["atmos", "all"]):
        averagers.latlon.xr_average(fyear, tar, modules)

    # -- Land
    # modules = {"land_month": "Land"}
    # if any(comp in comps for comp in ["land", "all"]):
    #    generic_driver(
    #        fyear,
    #        tar,
    #        modules,
    #        averagers.land_lm3.average,
    #        static_file=("land_static", "land_month"),
    #    )

    # -- Ice
    # modules = {"ice_month": "Ice"}
    # if any(comp in comps for comp in ["ice", "all"]):
    #    averagers.ice.xr_average(fyear, tar, modules)

    # -- Ocean
    modules = {
        "ocean_month": "Ocean",
    }
    if any(comp in comps for comp in ["ocean", "all"]):
        averagers.tripolar.xr_average(fyear, tar, modules)

    # -- OBGC
    modules = {
        "ocean_topaz_fluxes": "OBGC",
        "ocean_topaz_misc": "OBGC",
        "ocean_topaz_sfc_100": "OBGC",
        "ocean_topaz_tracers_month_z": "OBGC",
        "ocean_topaz_wc_btm": "OBGC",
    }
    if any(comp in comps for comp in ["obgc", "all"]):
        averagers.tripolar.xr_average(fyear, tar, modules)

    if any(comp in comps for comp in ["amoc"]):
        warnings.warn("AMOC calculation is not supported for ESM2.")

    # -- Close out the tarfile handle
    tar.close()



"""gfdlvitals - a package for computing global mean metrics"""

from .version import __version__

from . import averagers
from . import cli
from . import models
from . import sample
from . import util
from .extensions import *
from .plot import *

__all__ = [
    "averagers",
    "cli",
    "models",
    "util",
    "VitalsDataFrame",
    "open_db",
    "plot_timeseries",
]



""" Utiliites for use in averaging routines """

import numpy as np

from gfdlvitals.util.netcdf import extract_from_tar
from gfdlvitals.util.netcdf import tar_member_exists

import gfdlvitals.util.gmeantools as gmeantools
import gfdlvitals.util.netcdf as nctools

__all__ = ["RichVariable", "generic_cubesphere_driver", "process_var"]


class RichVariable:
    """Metadata-rich variable class

    Parameters
    ----------
    varname : str
        Variable name
    grid_file : io.BufferedReader or list of io.BufferedReader
        Grid-spec tiles
    data_file : io.BufferedReader or list of io.BufferedReader
        Data tiles
    fyear : str
        Year that is being processed
    outdir : str
        Output path directory
    label : str
        DB file name
    geolat : np.ma.masked_array
        Array of latitudes
    geolon : np.ma.masked_array
        Array of longitudes
    cell_area : np.ma.masked_array, optional
        Array of cell areas
    area_types : dict, optional
        [description], by default None
    cell_depth : np.ma.masked_array, optional
        [description], by default None
    cell_frac : np.ma.masked_array, optional
        [description], by default None
    soil_area : np.ma.masked_array, optional
        [description], by default None
    soil_frac : np.ma.masked_array, optional
        [description], by default None
    average_dt : np.ma.masked_array
        Array of time averaging period
    """

    def __init__(
        self,
        varname,
        grid_file,
        data_file,
        fyear,
        outdir,
        label,
        geolat,
        geolon,
        cell_area=None,
        area_types=None,
        cell_depth=None,
        cell_frac=None,
        soil_area=None,
        soil_frac=None,
        average_dt=None,
    ):
        self.varname = varname
        self.grid_file = grid_file
        self.data_file = data_file
        self.fyear = fyear
        self.outdir = outdir
        self.label = label
        self.geolat = geolat
        self.geolon = geolon
        self.cell_area = cell_area
        self.area_types = area_types
        self.cell_depth = cell_depth
        self.cell_frac = cell_frac
        self.soil_area = soil_area
        self.soil_frac = soil_frac
        self.average_dt = average_dt

    def __str__(self):
        return self.__class__.__name__

    def __hash__(self):
        return hash([self.__dict__[x] for x in list(self.__dict__.keys())])


def process_var(variable, averager=None):
    """Function called by multiprocessing thread to process a variable

    Parameters
    ----------
    variable : RichVariable object
        Input variable to process
    """

    var = None
    _cell_depth = None

    if averager in ["cubesphere", "land_lm4"]:
        # open up data tiles and get variable info
        data_file = [nctools.in_mem_nc(x) for x in variable.data_file]
        varshape = data_file[0].variables[variable.varname].shape
        ndim = len(varshape)
        units = gmeantools.extract_metadata(data_file[0], variable.varname, "units")
        long_name = gmeantools.extract_metadata(
            data_file[0], variable.varname, "long_name"
        )

        if (averager == "land_lm4") and (ndim >= 3):
            cell_measures = gmeantools.extract_metadata(
                data_file[0], variable.varname, "cell_measures"
            )
            area_measure = gmeantools.parse_cell_measures(cell_measures, "area")
            if (area_measure is not None) and (area_measure != "area_ntrl"):
                var = gmeantools.cube_sphere_aggregate(variable.varname, data_file)
                var = np.ma.average(
                    var, axis=0, weights=data_file[0].variables["average_DT"][:]
                )
                _area_weight = variable.area_types[area_measure]
                _cell_depth = variable.cell_depth if ndim == 4 else None

        elif (averager == "cubesphere") and (ndim == 3):
            var = gmeantools.cube_sphere_aggregate(variable.varname, data_file)
            var = np.ma.average(
                var, axis=0, weights=data_file[0].variables["average_DT"][:]
            )
            _area_weight = variable.cell_area

    else:
        fdata = nctools.in_mem_nc(variable.data_file)
        units = gmeantools.extract_metadata(fdata, variable.varname, "units")
        long_name = gmeantools.extract_metadata(fdata, variable.varname, "long_name")
        ndim = len(fdata.variables[variable.varname].shape)

        if (averager == "tripolar") and (ndim >= 3):
            dims = fdata.variables[variable.varname].dimensions
            if (dims[-2::] == ("yh", "xh")) and (ndim == 3):
                var = fdata.variables[variable.varname][:]
            elif (ndim == 4) and (variable.varname[0:9] == "tot_layer"):
                var = fdata.variables[variable.varname][:]
                var = np.ma.sum(var, axis=1).squeeze()
            if var is not None:
                var = np.ma.average(
                    var, axis=0, weights=fdata.variables["average_DT"][:]
                )
                _area_weight = variable.cell_area

        if averager == "ice":
            if ndim == len(variable.cell_area.shape):
                var = fdata.variables[variable.varname][:]

        if averager == "lat-lon" and ndim == 3:
            var = fdata.variables[variable.varname][:]
            var = np.ma.average(var, axis=0, weights=fdata.variables["average_DT"][:])
            _area_weight = variable.cell_area

        if averager == "land-lm3" and ndim >= 3:
            var = fdata[variable.varname][:]
            var = np.ma.average(var, axis=0, weights=fdata["average_DT"][:])

    if var is None:
        return None

    for reg in ["global", "tropics", "nh", "sh"]:
        sqlfile = (
            variable.outdir
            + "/"
            + variable.fyear
            + "."
            + reg
            + "Ave"
            + variable.label
            + ".db"
        )

        if averager == "ice":
            if var.shape != variable.cell_area.shape:
                return None

            # mask by latitude bands
            _v, _area = gmeantools.mask_latitude_bands(
                var, variable.cell_area, variable.geolat, region=reg
            )
            _v = np.ma.sum((_v * _area), axis=(-1, -2)) / np.ma.sum(
                _area, axis=(-1, -2)
            )

            # write out ice annual mean, min, and max
            gmeantools.write_sqlite_data(
                sqlfile, variable.varname + "_max", variable.fyear[:4], np.ma.max(_v)
            )
            gmeantools.write_sqlite_data(
                sqlfile, variable.varname + "_min", variable.fyear[:4], np.ma.min(_v)
            )
            gmeantools.write_sqlite_data(
                sqlfile,
                variable.varname + "_mean",
                variable.fyear[:4],
                np.ma.average(_v, axis=0, weights=variable.average_dt),
            )

            # sum of area not reported for ice diagnostics
            area_sum = None

        # -- Legacy LM3 Land Model
        elif averager == "land-lm3":
            result, summed = gmeantools.legacy_area_mean(
                var,
                variable.cell_area,
                variable.geolat,
                variable.geolon,
                cell_frac=variable.cell_frac,
                soil_frac=variable.soil_frac,
                region=reg,
                varname=variable.varname,
                component="land",
            )

            # use the legacy write method of the LM3 version of the land model
            gmeantools.write_sqlite_data(
                sqlfile,
                variable.varname,
                variable.fyear[:4],
                result,
                summed,
                component="land",
            )

            # legacy global mean format did not have land area
            area_sum = None

        else:
            if (averager == "land_lm4") and (_cell_depth is not None):
                if var.shape[0] != _cell_depth.shape[0]:
                    return None

            result, area_sum = gmeantools.area_mean(
                var,
                _area_weight,
                variable.geolat,
                variable.geolon,
                region=reg,
                cell_depth=_cell_depth,
            )

            if (averager == "land_lm4") and (hasattr(result, "mask")):
                return None

            gmeantools.write_sqlite_data(
                sqlfile, variable.varname, variable.fyear[:4], result
            )

        gmeantools.write_metadata(sqlfile, variable.varname, "units", units)
        gmeantools.write_metadata(sqlfile, variable.varname, "long_name", long_name)

        if averager == "land_lm4":
            area_measure = (
                area_measure.replace("area", "volume") if ndim == 4 else area_measure
            )

            gmeantools.write_metadata(
                sqlfile, variable.varname, "cell_measure", area_measure
            )

            gmeantools.write_sqlite_data(
                sqlfile, area_measure, variable.fyear[:4], area_sum
            )

        else:
            if area_sum is not None:
                gmeantools.write_sqlite_data(
                    sqlfile, "area", variable.fyear[:4], area_sum
                )

    if averager in ["cubesphere", "land_lm4"]:
        _ = [x.close() for x in data_file]
    else:
        fdata.close()

    return None


def generic_cubesphere_driver(fyear, tar, modules, average, grid_spec="grid_spec"):
    """Generic cubesphere data driver

    Parameters
    ----------
    fyear : str
        Year to process (YYYYMMDD)
    tar : tarfile object
        In-memory pointer to history tarfile
    modules : dict
        Dictionary of history nc streams (keys) and output db name (values)
    """

    members = [f"{fyear}.{x}.tile1.nc" for x in list(modules.keys())]
    members = [tar_member_exists(tar, x) for x in members]

    if any(members):
        gs_tiles = [
            extract_from_tar(tar, f"{fyear}.{grid_spec}.tile{x}.nc")
            for x in range(1, 7)
        ]

        for module in list(modules.keys()):
            if tar_member_exists(tar, f"{fyear}.{module}.tile1.nc"):
                print(f"{fyear} - {module}")
                data_tiles = [
                    extract_from_tar(tar, f"{fyear}.{module}.tile{x}.nc")
                    for x in range(1, 7)
                ]
                average(gs_tiles, data_tiles, fyear, "./", modules[module])
                del data_tiles

        del gs_tiles


def generic_driver(fyear, tar, modules, average, static_file=None):
    """Run the averager on tripolar ocean history data

    Parameters
    ----------
    fyear : str
        Year to process (YYYYMMDD)
    tar : tarfile object
        In-memory pointer to history tarfile
    modules : dict
        Dictionary of history nc streams (keys) and output db name (values)
    """
    members = [f"{fyear}.{x}.nc" for x in list(modules.keys())]
    members = [tar_member_exists(tar, x) for x in members]

    if any(members):

        if static_file is not None:

            assert isinstance(static_file, tuple), "Static file must be a tuple"
            assert (
                len(static_file) == 2
            ), "Static file tuple must have primary and one backup stream"

            staticname = f"{fyear}.{static_file[0]}.nc"
            grid_file = (
                extract_from_tar(tar, staticname)
                if tar_member_exists(tar, staticname)
                else extract_from_tar(tar, f"{fyear}.{static_file[1]}.nc")
            )

        else:
            grid_file = None

        for module in list(modules.keys()):
            fname = f"{fyear}.{module}.nc"
            if tar_member_exists(tar, fname):
                print(f"{fyear} - {module}")
                fdata = extract_from_tar(tar, fname)
                grid_file = fdata if grid_file is None else grid_file
                average(grid_file, fdata, fyear, "./", modules[module])
                del fdata

        del grid_file



""" Tools for working with xarray datasets """

import xarray as xr

from gfdlvitals.util.gmeantools import write_sqlite_data
from gfdlvitals.util.gmeantools import write_metadata

__all__ = ["xr_mask_by_latitude", "xr_to_db", "xr_weighted_avg"]


def xr_mask_by_latitude(arr, geolat, region=None):
    """Masks an xarray object based on a latitude range

    Parameters
    ----------
    arr : xarray.DataArray or xarray.DataSet
        Input unmasked object
    geolat : xarray.DataArray
        Data Array of latitude coordinates
    region : str, optional
        Predefined region of "global","nh","sh","tropics", by default None

    Returns
    -------
    Masked xarray data object
    """

    arr = arr.copy()

    if region == "nh":
        result = arr.where(geolat > 30.0, 0.0)
    elif region == "sh":
        result = arr.where(geolat < -30.0, 0.0)
    elif region == "tropics":
        result = arr.where((geolat >= -30.0) & (geolat <= 30.0), 0.0)
    else:
        result = arr

    return result


def xr_to_db(dset, fyear, sqlfile):
    """Writes Xarray dataset to SQLite format

    Parameters
    ----------
    dset : xarray.DataSet
        Input dataset
    fyear : str
        Year label (YYYY)
    sqlfile : str
        Filename of output db file
    """
    for var in list(dset.variables):
        write_sqlite_data(sqlfile, var, str(fyear), str(dset[var].data))
        if "units" in list(dset[var].attrs):
            write_metadata(sqlfile, var, "units", dset[var].units)
        if "long_name" in list(dset[var].attrs):
            write_metadata(sqlfile, var, "long_name", dset[var].long_name)
        if "measure" in list(dset[var].attrs):
            write_metadata(sqlfile, var, "cell_measure", dset[var].measure)


def xr_weighted_avg(dset, weights):
    """Generates weighted space and time average of an xarray DataSet

    Parameters
    ----------
    dset : xarray.DataSet
        Input dataset
    weights : xarray.DataArray or list
        Array to use for weights

    Returns
    -------
    xarray DataSet containing weighted averages
    """
    _weights = [weights] if not isinstance(weights, list) else weights

    _dset = xr.Dataset()
    result = xr.Dataset()

    for weight in _weights:
        variables = list(dset.variables.keys())
        for x in variables:
            if sorted(dset[x].dims) == sorted(weight.dims):
                _dset[x] = dset[x]

        _dset_weighted = _dset.weighted(weight).mean()
        for x in list(_dset_weighted.variables):
            _dset_weighted[x] = _dset_weighted[x].astype(dset[x].dtype)
            _dset_weighted[x].attrs = dset[x].attrs

        result = result.merge(_dset_weighted)

    return result



""" Ice model averaging routines """

import warnings

import numpy as np

import gfdlvitals.util.gmeantools as gmeantools
import gfdlvitals.util.xrtools as xrtools
import gfdlvitals.util.netcdf as netcdf


__all__ = ["xr_average"]


def xr_average(fyear, tar, modules):
    """xarray-based processing routines for lat-lon model output

    Parameters
    ----------
    fyear : str
        Year being processed (YYYY)
    tar : tarfile
        In-memory tarfile object
    modules : dict
        Mappings of netCDF file names inside the tar file to output db file names
    """

    members = [x for x in modules if netcdf.tar_member_exists(tar, f"{fyear}.{x}.nc")]

    for member in members:
        print(f"{fyear}.{member}.nc")
        data_file = netcdf.extract_from_tar(tar, f"{fyear}.ice_month.nc")
        dset = netcdf.in_mem_xr(data_file)

        if netcdf.tar_member_exists(tar, f"{fyear}.ice_static.nc"):
            grid_file = f"{fyear}.ice_static.nc"
        elif netcdf.tar_member_exists(tar, f"{fyear}.sea_ice_geometry.nc"):
            grid_file = f"{fyear}.sea_ice_geometry.nc"
        else:
            grid_file = f"{fyear}.ice_month.nc"

        grid_file = netcdf.extract_from_tar(tar, grid_file)
        ds_grid = netcdf.in_mem_xr(grid_file)

        # Retain only time-dependent variables
        variables = list(dset.variables.keys())
        for x in variables:
            if "time" not in dset[x].dims:
                del dset[x]
            if x == "CN":
                dset[x] = dset[x].sum(("ct")).assign_attrs(dset[x].attrs)

        if "CN" in list(dset.variables.keys()):
            concentration = dset["CN"]
        elif "siconc" in list(dset.variables.keys()):
            concentration = dset["siconc"]
        else:
            warnings.warn("Unable to determine sea ice concentation")

        if "Ah" in ds_grid.variables:
            ice_area_units = str(ds_grid.Ah.units)
            assert (
                ice_area_units == "m2"
            ), f"Ice area units {ice_area_units} are not correct. Expected m2"
            _area = ds_grid.Ah
            _area = _area.rename({"lath": "yT", "lonh": "xT"})
        else:
            earth_radius = 6371.0e3  # Radius of the Earth in 'm'
            if "CELL_AREA" not in ds_grid.variables:
                warnings.warn("Unable to find sea ice cell area. Skipping.")
                return

            _area = ds_grid["CELL_AREA"] * 4.0 * np.pi * (earth_radius ** 2)

        # --- todo Add in concentration and extent

        for region in ["global", "nh", "sh"]:

            if "geolat" in ds_grid.variables:
                _geolat = ds_grid["geolat"]
                _geolat = _geolat.rename({"lath": "yT", "lonh": "xT"})
            else:
                _geolat = ds_grid["GEOLAT"]

            _masked_area = xrtools.xr_mask_by_latitude(_area, _geolat, region=region)
            gmeantools.write_sqlite_data(
                f"{fyear}.{region}Ave{modules[member]}.db",
                "area",
                fyear,
                _masked_area.sum().data,
            )

            # area-weight but not time_weight
            weights = _masked_area
            _dset = dset.copy()

            ones = (concentration * 0.0) + 1.0
            ice_area = ones.where(concentration > 0.0, 0.0) * _masked_area
            extent = ones.where(concentration > 0.15, 0.0) * _masked_area

            ice_area_attrs = {
                "long_name": "area covered by sea ice",
                "units": "million km2",
            }
            extent_attrs = {"long_name": "sea ice extent", "units": "million km2"}

            for x in list(_dset.variables):
                if tuple(_dset[x].dims)[-3::] == ("time", "yT", "xT"):
                    _dset[x] = (
                        (_dset[x] * weights).sum(("yT", "xT")) / weights.sum()
                    ).assign_attrs(dset[x].attrs)
                    _dset["ice_area"] = (
                        ice_area.sum(("yT", "xT")) * 1.0e-12
                    ).assign_attrs(ice_area_attrs)
                    _dset["extent"] = (extent.sum(("yT", "xT")) * 1.0e-12).assign_attrs(
                        extent_attrs
                    )
                elif tuple(_dset[x].dims)[-3::] == ("time", "yt", "xt"):
                    _dset[x] = (
                        (_dset[x] * weights).sum(("yt", "xt")) / weights.sum()
                    ).assign_attrs(dset[x].attrs)
                    _dset["ice_area"] = (
                        ice_area.sum(("yt", "xt")) * 1.0e-12
                    ).assign_attrs(ice_area_attrs)
                    _dset["extent"] = (extent.sum(("yt", "xt")) * 1.0e-12).assign_attrs(
                        extent_attrs
                    )
                else:
                    del _dset[x]

            _dset_max = _dset.max(("time"))
            newvars = {x: x + "_max" for x in list(_dset_max.variables)}
            _dset_max = _dset_max.rename(newvars)

            _dset_min = _dset.min(("time"))
            newvars = {x: x + "_min" for x in list(_dset_min.variables)}
            _dset_min = _dset_min.rename(newvars)

            weights = dset.average_DT.astype("float")
            _dset_weighted = xrtools.xr_weighted_avg(_dset, weights)
            newvars = {x: x + "_mean" for x in list(_dset_weighted.variables)}
            _dset_weighted = _dset_weighted.rename(newvars)

            xrtools.xr_to_db(_dset_weighted, fyear, f"{fyear}.{region}AveIce.db")
            xrtools.xr_to_db(_dset_max, fyear, f"{fyear}.{region}AveIce.db")
            xrtools.xr_to_db(_dset_min, fyear, f"{fyear}.{region}AveIce.db")



""" Cubesphere averaging utilities """

import xarray as xr

import gfdlvitals.util.gmeantools as gmeantools
import gfdlvitals.util.xrtools as xrtools
import gfdlvitals.util.netcdf as netcdf

__all__ = ["xr_average"]


def xr_average(fyear, tar, modules):
    """xarray-based processing routines for cubed sphere atmos. output

    Parameters
    ----------
    fyear : str
        Year being processed (YYYY)
    tar : tarfile
        In-memory tarfile object
    modules : dict
        Mappings of netCDF file names inside the tar file to output db file names
    """

    members = [
        x for x in modules if netcdf.tar_member_exists(tar, f"{fyear}.{x}.tile1.nc")
    ]

    for member in members:
        print(f"{fyear}.{member}.nc")
        data_files = [
            netcdf.extract_from_tar(tar, f"{fyear}.{member}.tile{x}.nc")
            for x in range(1, 7)
        ]
        data_files = [netcdf.in_mem_xr(x) for x in data_files]
        dset = xr.concat(data_files, "tile")

        # Retain only time-dependent variables
        variables = list(dset.variables.keys())
        for x in variables:
            if "time" not in dset[x].dims:
                del dset[x]

        # Aggregate grid spec tiles
        grid_files = [
            netcdf.extract_from_tar(tar, f"{fyear}.grid_spec.tile{x}.nc")
            for x in range(1, 7)
        ]
        grid_files = [netcdf.in_mem_xr(x) for x in grid_files]
        ds_grid = xr.concat(grid_files, "tile")

        dset["area"] = ds_grid["area"]

        for region in ["global", "nh", "sh", "tropics"]:
            _masked_area = xrtools.xr_mask_by_latitude(
                dset.area, ds_grid.grid_latt, region=region
            )
            gmeantools.write_sqlite_data(
                f"{fyear}.{region}Ave{modules[member]}.db",
                "area",
                fyear,
                _masked_area.sum().data,
            )

            weights = dset.average_DT.astype("float") * _masked_area
            _dset_weighted = xrtools.xr_weighted_avg(dset, weights)
            xrtools.xr_to_db(
                _dset_weighted, fyear, f"{fyear}.{region}Ave{modules[member]}.db"
            )



"""Generic utilities module"""

from . import average
from . import extract_ocean_scalar
from . import git
from . import gmeantools
from . import merge
from . import netcdf
from . import xrtools

__all__ = [
    "average",
    "extract_ocean_scalar",
    "git",
    "gmeantools",
    "merge",
    "netcdf",
    "xrtools",
]



""" Driver for CM4 class models """

import os
import tarfile

from gfdlvitals import averagers
from gfdlvitals import diags
from gfdlvitals.util import extract_ocean_scalar
from gfdlvitals.util.netcdf import tar_member_exists

import gfdlvitals.util.netcdf as nctools


__all__ = ["routines"]


def routines(args, infile):
    """Driver routine for CM4-class models

    Parameters
    ----------
    args : argparse.parser
        Parsed commmand line arguments
    infile : str, pathlike
        History tar file path
    """

    # -- Open the tarfile
    tar = tarfile.open(infile)

    # -- Set the model year string
    fyear = str(infile.split("/")[-1].split(".")[0])
    print("Processing " + fyear)

    # -- Get list of components to process
    comps = args.component

    # -- Atmospheric Fields
    modules = {
        "atmos_month": "Atmos",
        "atmos_co2_month": "Atmos",
        "atmos_month_aer": "AtmosAer",
        "aerosol_month_cmip": "AeroCMIP",
    }
    if any(comp in comps for comp in ["atmos", "all"]):
        averagers.cubesphere.xr_average(fyear, tar, modules)

    # -- Land Fields
    modules = {"land_month": "Land"}
    if any(comp in comps for comp in ["land", "all"]):
        averagers.land_lm4.xr_average(fyear, tar, modules)

    # -- Ice
    modules = {"ice_month": "Ice"}
    if any(comp in comps for comp in ["ice", "all"]):
        averagers.ice.xr_average(fyear, tar, modules)

    # -- Ocean
    fname = f"{fyear}.ocean_scalar_annual.nc"
    if any(comp in comps for comp in ["ocean", "all"]):
        if tar_member_exists(tar, fname):
            print(f"{fyear}.ocean_scalar_annual.nc")
            fdata = nctools.extract_from_tar(tar, fname, ncfile=True)
            extract_ocean_scalar.mom6(fdata, fyear, "./")
            fdata.close()

    # -- OBGC
    modules = {
        "ocean_cobalt_sfc": "OBGC",
        "ocean_cobalt_misc": "OBGC",
        "ocean_cobalt_tracers_year": "OBGC",
        "ocean_cobalt_tracers_int": "OBGC",
        "ocean_bling": "OBGC",
        "ocean_bling_cmip6_omip_2d": "OBGC",
        "ocean_bling_cmip6_omip_rates_year_z": "OBGC",
        "ocean_bling_cmip6_omip_sfc": "OBGC",
        "ocean_bling_cmip6_omip_tracers_month_z": "OBGC",
        "ocean_bling_cmip6_omip_tracers_year_z": "OBGC",
    }
    if any(comp in comps for comp in ["obgc", "all"]):
        averagers.tripolar.xr_average(fyear, tar, modules)

    # -- AMOC
    if any(comp in comps for comp in ["amoc", "all"]):
        diags.amoc.mom6_amoc(fyear, tar)

    # -- Close out the tarfile handle
    tar.close()

    # -- Do performance timing
    #try:
    #    infile = infile.replace("/history/", "/ascii/")
    #    infile = infile.replace(".nc.tar", ".ascii_out.tar")
    #    label = "Timing"
    #    if os.path.exists(infile):
    #        diags.fms.timing(infile, fyear, "./", label)
    #except RuntimeError:
    #    pass



""" Pandas class extension for gfdlvitals """

import math
import sqlite3
import warnings

import cftime

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from scipy import stats

__all__ = [
    "VitalsDataFrame",
    "Timeseries",
    "open_db",
    "reformat_time_axis",
    "ttest_ind_auto",
]


def _autocorr(arr, lag=1):
    """Computes the sample autocorrelation function coeffficient (rho)
    for given lag

    Parameters
    ----------
    arr : numpy.array (1-d)
        Input data array
    lag : int, optional
        lag, by default 1

    Returns
    -------
    float
        lagged auto-correlation
    """

    flen = float(len(arr))
    ybar = float(sum(arr)) / flen
    denom = sum([(y - ybar) ** 2 for y in arr])
    numer = sum(
        [(y - ybar) * (ytpk - ybar) for (y, ytpk) in zip(arr[:-lag], arr[lag:])]
    )
    return numer / denom


def ttest_ind_auto(arr1, arr2, axis=0):
    """Performs a t-test that adjusts the degrees of freedom
    based on the autocorrelation of the dataset.
    See Krasting et al. 2013 for more details
    DOI: 10.1175/JCLI-D-12-00832.1

    Parameters
    ----------
    arr1 : numpy.ndarray
        First array of data points
    arr2 : numpy.ndarray
        Second array of data points
    axis : int, optional
        Axis to perform t-test, by default 0

    Returns
    -------
    tuple
        t-statistic, probability,
        arr1 autocorrelation, arr2 autocorrelation
    """
    arr1, arr2, axis = _chk2_asarray(arr1, arr2, axis)
    variance1 = np.var(arr1, axis, ddof=1)
    variance2 = np.var(arr2, axis, ddof=1)
    arrlen1 = arr1.shape[axis]
    arrlen2 = arr2.shape[axis]
    lag1r1 = _autocorr(np.split(arr1, arr1.shape[axis], axis))
    lag1r2 = _autocorr(np.split(arr2, arr2.shape[axis], axis))
    n1eff = arrlen1 * ((1 - lag1r1) / (1 + lag1r1))
    n2eff = arrlen2 * ((1 - lag1r2) / (1 + lag1r2))
    df = n1eff + n2eff - 2
    diff = np.mean(arr1, axis) - np.mean(arr2, axis)
    svar = ((arrlen1 - 1) * variance1 + (arrlen2 - 1) * variance2) / float(df)
    t = diff / np.sqrt(svar * (1.0 / arrlen1 + 1.0 / arrlen2))
    t = np.where((diff == 0) * (svar == 0), 1.0, t)  # define t=0/0 = 0, identical means
    prob = stats.distributions.t.sf(np.abs(t), df) * 2
    # use np.abs to get upper tail
    # distributions.t.sf currently does not propagate nans
    # this can be dropped, if distributions.t.sf propagates nans
    # if this is removed, then prob = prob[()] needs to be removed
    prob = np.where(np.isnan(t), np.nan, prob)
    if t.ndim == 0:
        t = t[()]
        prob = prob[()]
    result = [t, prob, lag1r1, lag1r2]
    result = [float(x) for x in result]
    return tuple(result)


def _chk2_asarray(arr1, arr2, axis):
    if axis is None:
        arr1 = np.ravel(arr1)
        arr2 = np.ravel(arr2)
        outaxis = 0
    else:
        arr1 = np.asarray(arr1)
        arr2 = np.asarray(arr2)
        outaxis = axis
    return arr1, arr2, outaxis


def _remove_trend(x, y, order=1, anomaly=True, return_coefs=False, coefs=None):
    """Internal function to remove a linear trend

    Parameters
    ----------
    x : numpy.ndarray
        independent axis
    y : np.ndarray
        dependent axis
    order : int, optional
        Polynomial order to use for fitting, by default 1
    anomaly : bool, optional
        Results as anomalies from the fit, by default True
    return_coefs : bool, optional
        Return polynomial fit coeffiecients, by default False
    coefs : tuple, optional
        Use a provided set of coefficients, by default None

    Returns
    -------
    numpy.ndarray
        Detrended data
    """

    if None in list(y):
        return None
    if coefs is None:
        idx = np.isfinite(x) & np.isfinite(y)
        coefs = np.polyfit(x[idx], y[idx], order)
        if return_coefs is True:
            return coefs
    model = np.poly1d(coefs)
    fit = model(x)
    if anomaly is True:
        result = y - fit
    else:
        fit = fit - fit[0]
        result = y - fit
    return result


def _calc_trend(x, y, order=1):
    """Internal function to calculate trend line/curves

    Parameters
    ----------
    x : numpy.ndarray
        independent axis
    y : np.ndarray
        dependent axis
    order : int, optional
        Polynomial order to use for fitting, by default 1

    Returns
    -------
    numpy.ndarray
        Polynomial fit
    """

    idx = np.isfinite(x) & np.isfinite(y)
    coefs = np.polyfit(x[idx], y[idx], order)
    model = np.poly1d(coefs)
    return model(x)


def _remove_reference_trend(t, x, other, anomaly=True):
    """Removes trends from a reference dataset

    Parameters
    ----------
    t : numpy.ndarray
        independent axis
    x : gfdlvitals.VitalsDataFrame
        Dataset to be detrended
    other : gfdlvitals.VitalsDataFrame
        Reference dataset containing the trend to be removed
    anomaly : bool, optional
        Results as anomalies from the fit, by default True

    Returns
    -------
    gfdlvitals.VitalsDataFrame
        Detrended dataset
    """

    if str(x.name) not in list(other.columns):
        result = None
    else:
        _coefs = other[x.name].to_numpy()
        result = _remove_trend(t, x, anomaly=anomaly, coefs=_coefs)
    return result


def reformat_time_axis(ax=None):
    """Reformats x-axis labels to YYYY format

    Parameters
    ----------
    ax : matplotlob.pyplot.figure.axis, optional
        axis object to be reformatted, by default None
    """

    if ax is None:
        ax = plt.gca()
    labels = [x.get_text() for x in ax.xaxis.get_ticklabels()]
    labels = [x.split("-")[0] for x in labels]
    _ = ax.set_xticklabels(labels)


class VitalsDataFrame(pd.DataFrame):
    """Pandas class extension for holding the Vitals data

    Parameters
    ----------
    pd : pandas.DataFrame
        Input Pandas DataFrame object

    Returns
    -------
    gfdlvitals.VitalsDataFrame
        Class extension object
    """

    # temporary properties
    _internal_names = pd.DataFrame._internal_names + ["internal_cache"]
    _internal_names_set = set(_internal_names)

    # normal properties
    _metadata = ["added_property"]

    @property
    def _constructor(self):
        return VitalsDataFrame

    def ttest(self, df2):
        """Performs t-test between two instances of VitalsDataFrame

        Parameters
        ----------
        df2 : VitalsDataFrame
            Comparison data set

        Returns
        -------
        pandas.DataFrame
            Contains p-values for variables common between the
            two VitalsDataFrames
        """

        # get a list of common variables
        varlist = list(set(self.columns).intersection(df2.columns))

        # ignore some known fields
        ignore_list = ["area"]
        varlist = [x for x in varlist if x not in ignore_list]

        # perform t-test
        result = {k: ttest_ind_auto(self[k], df2[k])[1] for k in varlist}

        return pd.DataFrame(result, index=["pval"]).transpose()

    def smooth(self, window, extrap=False):
        """Apply a smoother to the dataset

        Parameters
        ----------
        window : int
            Smoothing filter length
        extrap : bool, optional
            Extrapolate data on the ends, by default False

        Returns
        -------
        self
            Smoothed dataset
        """

        if window is None:
            _df = self
        else:
            _df = self.rolling(window, center=True).mean()
            if extrap is True:
                _df.fillna(method="ffill", inplace=True)
                _df.fillna(method="bfill", inplace=True)
        _df.attrs = self.attrs
        return _df

    def extend(self, maxlen):
        """Extend VitalsDataFrame to a set length and pad with NaNs

        Parameters
        ----------
        maxlen : int
            New length of the VitalsDataFrame

        Returns
        -------
        self
            Extended dataset
        """
        endyear = tuple(self.index[-1].timetuple())
        padding = np.arange(1, maxlen - len(self.index) + 1) + endyear[0]
        added_index = [cftime.DatetimeNoLeap(x, *endyear[1:]) for x in padding]
        _df = pd.DataFrame({"times": added_index}).set_index("times")
        _df = self.append(_df)
        _df.attrs = self.attrs
        for column in self.columns:
            _df[column].attrs = self[column].attrs
        return _df

    def detrend(self, reference=None, order=1, anomaly=True, return_coefs=False):
        """Detrend VitalsDataFrame object

        Parameters
        ----------
        reference : gfdlvitals.DataFrame, optional
            Reference VitalsDataFrame, by default None
        order : int, optional
            Polynomial order to use for fitting, by default 1
        anomaly : bool, optional
            Results as anomalies from the fit, by default True
        return_coefs : bool, optional
            Return polynomial fit coeffiecients, by default False

        Returns
        -------
        self
            Extended dataset
        """
        tindex = np.array(
            [
                cftime.date2num(x, "days since 0001-01-01", calendar="noleap")
                for x in self.index
            ]
        )
        if reference is not None:
            if order != 1:
                print(
                    "Only a linear trend can be removed based on "
                    + "another dataset. Setting order to 1."
                )
                order = 1
            coefs = reference.detrend(order=order, return_coefs=True)
            result = self.apply(
                lambda x: (_remove_reference_trend(tindex, x, coefs, anomaly=anomaly))
            )
        else:
            result = self.apply(
                lambda x: (
                    _remove_trend(
                        tindex,
                        x,
                        order=order,
                        anomaly=anomaly,
                        return_coefs=return_coefs,
                    )
                )
            )
        result.attrs = self.attrs
        return result

    def trend(self, order=1):
        """Fits a trend to the VitalsDataFrame object

        Parameters
        ----------
        order : int, optional
            Polynomial order to use for fitting, by default 1

        Returns
        -------
        self
            Fitted trend dataset
        """
        tindex = np.array(
            [
                cftime.date2num(x, "days since 0001-01-01", calendar="noleap")
                for x in self.index
            ]
        )
        result = self.apply(lambda x: (_calc_trend(tindex, x, order=order)))
        result.attrs = self.attrs
        return result


class Timeseries:
    """Timeseries class object

    Parameters
    ----------
    f : str, path-like
        Input SQLite file
    var : str
        Variable to extract
    scale : float, optional
        Scale data by this factor, by default 1.0
    multiply_by_area : bool, optional
        Multiply variable by cell area before returning, by default False
    legacy_land : bool, optional
        Read legacy version of the land SQLite files, by default False
    start : int, optional
        Specify start year, by default None
    end : int, optional
        Specify end year, by default None
    """

    def __init__(
        self,
        f,
        var,
        scale=1.0,
        multiply_by_area=False,
        legacy_land=False,
        start=None,
        end=None,
    ):

        # open the sqlite connection
        con = sqlite3.connect(f)
        cur = con.cursor()

        # get the results from the sqlite file for the requested variable
        if legacy_land is True:
            _ = cur.execute("SELECT year,sum FROM " + var)
        else:
            _ = cur.execute("SELECT year,value FROM " + var)
        results = cur.fetchall()

        # split list of tuples into two different lists
        self._t, self._data = zip(*results)
        self._t = np.array(self._t)

        _ = cur.execute("SELECT name FROM sqlite_master where TYPE='table'")
        tables = [str(record[0]) for record in cur.fetchall()]

        if multiply_by_area is True:
            if "cell_measure" in tables:
                _ = cur.execute(
                    "SELECT value FROM cell_measure where var='" + var + "'"
                )
                cell_measure = cur.fetchone()[0]
            else:
                cell_measure = "area"
            _ = cur.execute("SELECT value FROM " + cell_measure)
            area = np.array(cur.fetchall()).squeeze()
            scale = area * scale

        self._data = np.array(self._data) * scale

        if "long_name" in tables:
            _ = cur.execute(f"SELECT value FROM long_name where var='{var}'")
            result = cur.fetchone()
            self.long_name = result[0] if isinstance(result, tuple) else None
        else:
            self.long_name = None

        if "units" in tables:
            _ = cur.execute(f"SELECT value FROM units where var='{var}'")
            result = cur.fetchone()
            self.units = result[0] if isinstance(result, tuple) else None
        else:
            self.units = None

        # close the connection
        cur.close()
        con.close()

        # filter based on start year and end year
        if start is not None:
            idx = [i for i, val in enumerate(self._t) if val >= start]
            self._t = self._t[idx]
            self._data = self._data[idx]
        else:
            start = self._t.min()
        if end is not None:
            idx = [i for i, val in enumerate(self._t) if val <= end]
            self._t = self._t[idx]
            self._data = self._data[idx]
        else:
            end = self._t.max() + 1

        # check for missing values and pad with nans
        missing_times = set(np.arange(start, end)) - set(self._t)
        if len(list(missing_times)) != 0:
            warnings.warn(f"Timeseries is incomplete for {var}: {missing_times}")

        # pad missing values with nans
        self.dict = dict(zip(self._t, self._data))
        self.dict = {
            **self.dict,
            **dict(zip(missing_times, [np.nan] * len(missing_times))),
        }

    @property
    def t(self):
        k, v = zip(*sorted(self.dict.items()))
        return k

    @property
    def data(self):
        k, v = zip(*sorted(self.dict.items()))
        return v

    def __str__(self):
        return self.__class__.__name__

    def __hash__(self):
        return hash([self.__dict__[x] for x in list(self.__dict__.keys())])


def open_db(
    dbfile,
    variables=None,
    yearshift=0.0,
    legacy_land=False,
    start=None,
    end=None,
):
    """Function to read sqlite dbfile"""

    if variables is None:
        conn = sqlite3.connect(dbfile)
        cur = conn.cursor()
        sql = "SELECT name FROM sqlite_master WHERE type='table'"
        _ = cur.execute(sql)
        variables = [str(record[0]) for record in cur.fetchall()]
        cur.close()
        conn.close()
        removes = ["units", "long_name", "cell_measure"]
        variables = [x for x in variables if x not in removes]

    # -- Loop over variables
    data = {}
    attributes = {}
    years = []
    skipped = []
    for var in variables:
        tsobj = Timeseries(dbfile, var, legacy_land=legacy_land, start=start, end=end)
        if len(tsobj.t) > 0:
            data[var] = tsobj.data
            years = years + list(tsobj.t)
            attributes[var] = {"long_name": tsobj.long_name, "units": tsobj.units}

    years = list(set(years))
    years = [x + float(yearshift) for x in years]

    variables = list(set(variables) - set(skipped))

    if start is None:
        start = -1 * math.inf

    if end is None:
        end = math.inf

    df = pd.DataFrame(data, index=years)
    df = df[(df.index >= start) & (df.index <= end)]
    df.index = cftime.num2date(
        (df.index * 365.0) - (365.0 / 2.0) - 1,
        "days since 0001-01-01",
        calendar="365_day",
    )

    df = VitalsDataFrame(df)
    df = df.sort_index()

    for var in list(df.columns):
        df[var].attrs = attributes[var]

    return df



""" Extract scalar fields from ocean model output """

from . import gmeantools

__all__ = ["mom6"]


def mom6(fdata, fyear, outdir):
    """Extract MOM6 scalar output and save to sqlite

    Parameters
    ----------
    fdata : netCDF4.Dataset
        Input ocean_scalar.nc file
    fyear : str
        Year being processed
    outdir : str
        Path to output directory
    """

    ignore_list = ["time_bounds", "time_bnds", "average_T2", "average_T1", "average_DT"]

    var_dict = fdata.variables.keys()
    var_dict = list(set(var_dict) - set(ignore_list))

    for varname in var_dict:
        if len(fdata.variables[varname].shape) == 2:
            units = gmeantools.extract_metadata(fdata, varname, "units")
            long_name = gmeantools.extract_metadata(fdata, varname, "long_name")
            result = fdata.variables[varname][0, 0]
            sqlfile = outdir + "/" + fyear + ".globalAveOcean.db"
            gmeantools.write_metadata(sqlfile, varname, "units", units)
            gmeantools.write_metadata(sqlfile, varname, "long_name", long_name)
            gmeantools.write_sqlite_data(sqlfile, varname, fyear[:4], result)



"""Land LM4.1 Averaging Routines"""

import xarray as xr

import gfdlvitals.util.gmeantools as gmeantools
import gfdlvitals.util.xrtools as xrtools
import gfdlvitals.util.netcdf as netcdf


__all__ = ["xr_average"]


def xr_average(fyear, tar, modules):
    """xarray-based processing routines for cubed sphere LM4 land output

    Parameters
    ----------
    fyear : str
        Year being processed (YYYY)
    tar : tarfile
        In-memory tarfile object
    modules : dict
        Mappings of netCDF file names inside the tar file to output db file names
    """

    members = [
        x for x in modules if netcdf.tar_member_exists(tar, f"{fyear}.{x}.tile1.nc")
    ]

    for member in members:
        print(f"{fyear}.{member}.nc")
        data_files = [
            netcdf.extract_from_tar(tar, f"{fyear}.{member}.tile{x}.nc")
            for x in range(1, 7)
        ]
        data_files = [netcdf.in_mem_xr(x) for x in data_files]
        dset = xr.concat(data_files, "tile")

        # Calculate cell depth
        depth = dset["zhalf_soil"].data
        depth = [depth[x] - depth[x - 1] for x in range(1, len(depth))]
        dset["depth"] = xr.DataArray(depth, dims=("zfull_soil"))
        depth = dset["depth"]

        # Retain only time-dependent variables
        variables = list(dset.variables.keys())
        for x in variables:
            if "time" not in dset[x].dims:
                del dset[x]

        # Load grid data
        grid_files = [
            netcdf.extract_from_tar(tar, f"{fyear}.land_static.tile{x}.nc")
            for x in range(1, 7)
        ]
        grid_files = [netcdf.in_mem_xr(x) for x in grid_files]
        ds_grid = xr.concat(grid_files, "tile")

        # Retain only time-invariant area fields
        grid = xr.Dataset()
        variables = list(ds_grid.variables.keys())
        for x in variables:
            if "area" in x or "frac" in x:
                grid[x] = ds_grid[x]

        # Get List of cell measures
        cell_measures = [
            dset[x].attrs["cell_measures"]
            for x in list(dset.variables)
            if "cell_measures" in list(dset[x].attrs.keys())
        ]
        cell_measures = sorted(list(set(cell_measures)))

        # Create dict of land groups based on cell measures
        land_groups = {}
        for x in cell_measures:
            land_groups[x] = xr.Dataset()

        # Loop over variables and assign them to groups
        variables = list(dset.variables.keys())
        for x in variables:
            if "cell_measures" in list(dset[x].attrs.keys()):
                _measure = dset[x].attrs["cell_measures"]
                dset[x].attrs["measure"] = _measure.split(" ")[-1]
                land_groups[_measure][x] = dset[x]

        # Since natural tile area is time-dependent, ignore for now
        if "area: area_ntrl" in cell_measures:
            cell_measures.remove("area: area_ntrl")

        if "area: glac_area" in cell_measures:
            cell_measures.remove("area: glac_area")

        # Loop over groups
        for measure in cell_measures:
            _dset = land_groups[measure]

            _measure = measure.split(" ")[-1]
            _area = ds_grid[_measure]

            for region in ["global", "nh", "sh", "tropics"]:
                _masked_area = xrtools.xr_mask_by_latitude(
                    _area, ds_grid.geolat_t, region=region
                )
                gmeantools.write_sqlite_data(
                    f"{fyear}.{region}Ave{modules[member]}.db",
                    _measure,
                    fyear,
                    _masked_area.sum().data,
                )

                # _masked_area = _masked_area.fillna(0)

                weights = dset.average_DT.astype("float") * _masked_area
                if _measure == "soil_area":
                    area_x_depth = _masked_area * depth
                    gmeantools.write_sqlite_data(
                        f"{fyear}.{region}Ave{modules[member]}.db",
                        "soil_volume",
                        fyear,
                        area_x_depth.sum().data,
                    )
                    weights = [
                        weights,
                        (weights * depth).transpose(
                            "tile", "time", "zfull_soil", "grid_yt", "grid_xt"
                        ),
                    ]
                    for x in list(_dset.variables):
                        if "zfull_soil" in list(_dset[x].dims):
                            _dset[x].attrs["measure"] = "soil_volume"

                _dset_weighted = xrtools.xr_weighted_avg(_dset, weights)

                xrtools.xr_to_db(
                    _dset_weighted, fyear, f"{fyear}.{region}Ave{modules[member]}.db"
                )



""" Module for working with top-level git repository """

import os
import subprocess
import sys

__all__ = ["is_clean", "retrieve_commit"]


def is_clean(path):
    """Determines if a git repository is clean

    Parameters
    ----------
    path : str, path-like
        Path to git repository

    Returns
    -------
    bool
        True if clean, False if modified
    """
    retcode = _status(path)
    return retcode == 0


def retrieve_commit(path):
    """Returns current git commit

    Parameters
    ----------
    path : str, path-like
        Path to git repository

    Returns
    -------
    str
        Git commit hash
    """
    cwd = os.getcwd()
    os.chdir(path)
    cmd = "git rev-parse HEAD"
    commit = subprocess.check_output(cmd.split(" "))
    commit = str(commit.decode("utf-8")).replace("\n", "")
    os.chdir(cwd)
    return commit


def _status(path):
    """Internal subprocess call to query status of git repo

    Parameters
    ----------
    path : str, path-like
        Path to git repository

    Returns
    -------
    int
        Return code of system call to `git diff-index`
    """
    cwd = os.getcwd()
    os.chdir(path)
    cmd = "git diff-index --quiet HEAD"
    result = subprocess.run(cmd.split(" "), stdout=subprocess.DEVNULL, check=False)
    os.chdir(cwd)
    return result.returncode


if __name__ == "__main__":
    PATH = os.path.abspath(sys.argv[1]) if len(sys.argv) > 1 else os.getcwd()
    COMMIT = retrieve_commit(PATH)
    print(COMMIT)
    RETCODE = _status(PATH)
    print(RETCODE)
    CLEAN = is_clean(PATH)
    print(CLEAN)



""" Sample db files for demonstration """

import pkg_resources as pkgr

historical = pkgr.resource_filename("gfdlvitals", "resources/historical.db")
picontrol = pkgr.resource_filename("gfdlvitals", "resources/picontrol.db")
